import platform
import sys

from easydict import EasyDict as edict
import numpy as np
from qtpy import QtWidgets, QtCore

from pymodaq.control_modules.viewer_utility_classes import DAQ_Viewer_base
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.control_modules.viewer_utility_classes import comon_parameters
from pymodaq.utils.data import DataFromPlugins, Axis
from pymodaq.utils.daq_utils import ThreadCommand, find_dict_in_list_from_key_val, zeros_aligned
from pymodaq.utils.parameter.utils import iter_children

arch, plat = platform.architecture()
from time import perf_counter
logger = set_logger(get_module_name(__file__))
libpath = '/'

if plat.startswith('Windows'):
    libpath = 'C:\\Program Files\\Andor SDK3\\win32'
    if libpath not in sys.path:
        sys.path.append(libpath)

try:
    from pymodaq_plugins_andor.hardware.andor_sdk3 import api, sdk3cam
except OSError as e:
    logger.exception(e)

try:
    models, names, serial_numbers = api.getCameraInfos()
    camera_list = [f'{models[ind]} {serial_numbers[ind]}' for ind in range(len(models))]
except Exception as e:
    logger.exception(f'Impossible to communicate with camera, try to set another library path than {libpath}')
    camera_list = []


class DAQ_2DViewer_AndorSCMOS(DAQ_Viewer_base):
    """
        Base class for Andor SCMOS camera


        =============== ==================
        **Attributes**   **Type**

        =============== ==================

        See Also
        --------
        utility_classes.DAQ_Viewer_base
    """
    start_waitloop = QtCore.Signal(int, int)
    stop_waitloop = QtCore.Signal()

    hardware_averaging = True  # will use the accumulate acquisition mode if averaging is neccessary
    live_mode_available = True
    params = comon_parameters + [
        {'title': 'Dll library:', 'name': 'andor_lib', 'type': 'browsepath', 'value': libpath},

        {'title': 'Camera Settings:', 'name': 'camera_settings', 'type': 'group', 'children': [
            {'title': 'Camera Models:', 'name': 'camera_model', 'type': 'list', 'limits': camera_list},
            {'title': 'Exposure (ms):', 'name': 'exposure', 'type': 'float', 'value': 0.01, 'default': 0.01, 'min': 0},
            {'title': 'Frame Rate (Hz):', 'name': 'frame_rate', 'type': 'float', 'value': 0., 'readonly': True},
            {'title': 'Number of buffer:', 'name': 'buffer_number', 'type': 'int', 'value': 10, 'min': 1},
            {'title': '', 'name': 'reset_buffers', 'type': 'bool_push', 'value': False, 'label': 'Reset Buffers'},

            {'title': 'Sensor size:', 'name': 'image_size', 'type': 'group', 'children': [
                {'title': 'Nx:', 'name': 'Nx', 'type': 'int', 'value': 0, 'default': 0, 'readonly': True},
                {'title': 'Ny:', 'name': 'Ny', 'type': 'int', 'value': 0, 'default': 0, 'readonly': True},
                ]},
            {'title': 'Encoding:', 'name': 'encoding', 'type': 'list', 'limits': []},
            {'title': 'Image Area:', 'name': 'image_settings', 'type': 'group', 'children': [
                {'title': 'Binning:', 'name': 'binning', 'type': 'list', 'limits': []},
                {'title': 'Binning along x:', 'name': 'bin_x', 'type': 'int', 'value': 1, 'default': 1, 'min': 1},
                {'title': 'Binning along y:', 'name': 'bin_y', 'type': 'int', 'value': 1, 'default': 1, 'min': 1},
                {'title': 'Left:', 'name': 'im_left', 'type': 'int', 'value': 1, 'default': 1, 'min': 1},
                {'title': 'Top:', 'name': 'im_top', 'type': 'int', 'value': 1024, 'default': 1024, 'min': 1},
                {'title': 'Width (in binned pxls):', 'name': 'im_width', 'type': 'int', 'value': 1024, 'min': 1},
                {'title': 'Height (in binned pxls):', 'name': 'im_height', 'type': 'int', 'value': 256, 'min': 1, },
                {'title': 'Set Size Max:', 'name': 'max_size', 'type': 'bool_push', 'value': False,
                 'label': 'Max'},
                {'title': 'Readout time (ms):', 'name': 'readout_time', 'type': 'float', 'value': 0., 'readonly': True},
                ]},

            {'title': 'Trigger Settings:', 'name': 'trigger', 'type': 'group', 'children': [
                {'title': 'Mode:', 'name': 'trigger_mode', 'type': 'list', 'limits': []},
                {'title': 'Software Trigger:', 'name': 'soft_trigger', 'type': 'bool_push', 'value': False,
                 'label': 'Fire'},
                {'title': 'External Trigger delay (ms):', 'name': 'ext_trigger_delay', 'type': 'float', 'value': 0.},
            ]},

            {'title': 'Shutter Settings:', 'name': 'shutter', 'type': 'group', 'children': [
                {'title': 'Mode:', 'name': 'shutter_mode', 'type': 'list', 'limits': []},
                {'title': 'External Trigger is:', 'name': 'shutter_on_ext_trigger', 'type': 'list', 'limits': []},
                ]},


            {'title': 'Temperature Settings:', 'name': 'temperature_settings', 'type': 'group', 'children': [
                {'title': 'Enable Cooling:', 'name': 'enable_cooling', 'type': 'bool', 'value': True},
                {'title': 'Set Point:', 'name': 'set_point', 'type': 'list', 'limits': []},
                {'title': 'Current value:', 'name': 'current_value', 'type': 'float', 'value': 20, 'readonly': True},
                {'title': 'Status:', 'name': 'status', 'type': 'list', 'limits': [], 'readonly': True},
                ]},
            ]},
        ]

    def ini_attributes(self):

        self.camera_controller: api.AndorCamera = None

        self.buffers = []
        self.buffers_pointer = []
        self._Nbuffers = None
        self._reset_buffers_cmd = False
        self.refresh_time_fr = 200

        self.current_buffer = -1
        self.n_grabed_data = None
        self.n_grabed_frame_rate = None
        self.start_time = None
        self.live = False
        self.wait_time = 0

        self.x_axis = None
        self.y_axis = None
        self.camera_controller = None
        self.data = None
        self.SIZEX, self.SIZEY = (None, None)
        self.camera_done = False
        self.acquirred_image = None
        self.callback_thread = None
        self.Naverage = None
        self.data_shape = None  # 'Data2D' if sizey != 1 else 'Data1D'

        self.temperature_timer = QtCore.QTimer()
        self.temperature_timer.timeout.connect(self.update_temperature)

    def commit_settings(self, param):
        """
            | Activate parameters changes on the hardware from parameter's name.
            |

            =============== ================================    =========================
            **Parameters**   **Type**                           **Description**
            *param*          instance of pyqtgraph parameter    The parameter to activate
            =============== ================================    =========================

            Three profile of parameter :
                * **bin_x** : set binning camera from bin_x parameter's value
                * **bin_y** : set binning camera from bin_y parameter's value
                * **set_point** : Set the camera's temperature from parameter's value.

        """
        try:
            self.stop()
            QtWidgets.QApplication.processEvents()

            if param.name() == 'set_point':
                self.camera_controller.SetTemperature(param.value())

            elif param.name() == 'reset_buffers':
                self._reset_buffers_cmd = True

            elif param.name() == 'exposure':
                self.camera_controller.ExposureTime.setValue(self.settings.child('camera_settings', 'exposure').value() * 1e-3)
                self.settings.child('camera_settings', 'exposure').setValue(
                    self.camera_controller.ExposureTime.getValue() * 1e3)

            elif param.name() == 'encoding':
                self.camera_controller.SimplePreAmpGainControl.setString(param.value())

            elif param.name() in iter_children(self.settings.child('camera_settings', 'shutter'), []):
                self.set_shutter()

            elif param.name() in iter_children(self.settings.child('camera_settings', 'image_settings'), []):
                if param.name() == 'binning': #if binning set from the preselection, update first the binx and biny values before setting the image area
                    self.camera_controller.AOIBinning.setString(param.value())
                    self.settings.child('camera_settings', 'image_settings', 'bin_x').setValue(
                        self.camera_controller.AOIHBin.getValue())
                    self.settings.child('camera_settings', 'image_settings', 'bin_y').setValue(
                        self.camera_controller.AOIVBin.getValue())
                    QtWidgets.QApplication.processEvents()

                if param.name() == 'bin_x':
                    self.camera_controller.AOIHBin.setValue(param.value())
                    # self.settings.child('camera_settings', 'image_settings', 'im_width').setValue(
                    #     self.controller.AOIWidth.getValue())
                    QtWidgets.QApplication.processEvents()

                if param.name() == 'bin_y':
                    self.camera_controller.AOIVBin.setValue(param.value())
                    # self.settings.child('camera_settings', 'image_settings', 'im_height').setValue(
                    #     self.controller.AOIHeight.getValue())
                    QtWidgets.QApplication.processEvents()

                if param.name() == 'max_size' and param.value():
                    #param.setValue(False)
                    #self.setup_image()
                    self.settings.child('camera_settings', 'image_settings', 'im_width').setValue(
                        self.camera_controller.AOIWidth.max())
                    # self.send_param_status(self.settings.child('camera_settings', 'image_settings', 'im_width'),
                    #                        [(self.settings.child('camera_settings', 'image_settings', 'im_width'),
                    #                         'value', self.controller.AOIWidth.max())])

                    self.settings.child('camera_settings', 'image_settings', 'im_height').setValue(
                        self.camera_controller.AOIHeight.max())

                self.setup_image()
                QtWidgets.QApplication.processEvents()
                #self.set_image_area()

            elif param.name() in iter_children(self.settings.child('camera_settings', 'temperature_settings'), []):
                self.setup_temperature()

            elif param.name() == 'soft_trigger':
                if param.value():
                    self.camera_controller.SoftwareTrigger()
                    param.setValue(False)

            elif param.name() in iter_children(self.settings.child('camera_settings', 'trigger'), []):
                self.set_trigger()

        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), 'log']))

    def set_trigger(self):
        self.camera_controller.TriggerMode.setString(self.settings.child('camera_settings',
                                                                         'trigger', 'trigger_mode').value())
        if 'External' in self.camera_controller.TriggerMode.getString():
            self.settings.child('camera_settings', 'trigger',
                                'ext_trigger_delay').setLimits((self.camera_controller.ExternalTriggerDelay.min(),
                                                                self.camera_controller.ExternalTriggerDelay.max()))

            self.camera_controller.ExternalTriggerDelay.setValue(
                self.settings.child('camera_settings', 'trigger', 'ext_trigger_delay').value() / 1000)
            self.settings.child('camera_settings', 'trigger',
                                'ext_trigger_delay').setValue(self.camera_controller.ExternalTriggerDelay.getValue() * 1000)

    def emit_data(self, buffer_pointer):
        """
            Fonction used to emit data obtained by callback.

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            buff_temp = buffer_pointer[0]
            self.current_buffer += 1
            self.n_grabed_data += 1
            self.n_grabed_frame_rate += 1
            #print(f'ind_grabemit:{self.n_grabed_data}')
            self.current_buffer = self.current_buffer % self._Nbuffers
            #print(f'ind_current_buffer:{self.current_buffer}')

            if self.buffers[self.current_buffer].ctypes.data != buff_temp:
                # buff_val = [self.buffers[ind].ctypes.data for ind in range(len(self.buffers))].index(buff_temp)
                # print(f'Buffer index should be {self.current_buffer} but is in fact {buff_val}')
                self.stop()
                QtWidgets.QApplication.processEvents()

                self.emit_status(ThreadCommand('Update_Status',
                                               ['Returned buffer not equal to expected buffer,'
                                                ' restarting acquisition and'
                                                ' freeing buffers', 'log']))
                logger.warning('Returned buffer not equal to expected buffer, restarting acquisition and'
                               ' freeing buffers')
                self._reset_buffers_cmd = True
                self.grab_data(self.Naverage, live=self.live, wait_time=self.wait_time)
                return

            cam_name = self.settings.child('camera_settings', 'camera_model').value()
            Nx = self.settings.child('camera_settings', 'image_settings', 'im_width').value()
            Ny = self.settings.child('camera_settings', 'image_settings', 'im_height').value()
            data = self.camera_controller.get_image_fom_buffer(Nx, Ny, self.buffers[self.current_buffer]).T

            if self.n_grabed_data % self.Naverage == 0 and self.live:
                self.data = 1 / self.Naverage * data
            else:
                self.data += 1 / self.Naverage * data

            if not self.live:
                if self.n_grabed_data > self.Naverage:
                    self.stop()
                else:
                    #self.data += 1 / self.Naverage * data
                    if self.n_grabed_data == self.Naverage:
                        self.data_grabed_signal.emit([
                            DataFromPlugins(name=cam_name, data=[self.data], dim=self.data_shape)])

                    elif self.n_grabed_data < self.Naverage:
                        self.data_grabed_signal_temp.emit([
                            DataFromPlugins(name=cam_name, data=[self.data * self.Naverage / self.n_grabed_data],
                                            dim=self.data_shape)])
            else:  # in live mode
                if perf_counter() - self.start_time > self.refresh_time_fr / 1000:  # refresh the frame rate every
                    # refresh_time_fr ms
                    self.settings.child('camera_settings',
                                        'frame_rate').setValue(self.n_grabed_frame_rate / (self.refresh_time_fr / 1000))
                    self.start_time = perf_counter()
                    self.n_grabed_frame_rate = 0

                if self.n_grabed_data % self.Naverage == 0:
                    self.data_grabed_signal.emit([
                        DataFromPlugins(name=cam_name, data=[self.data], dim=self.data_shape)])
                else:
                    if self.n_grabed_data % self.Naverage != 0:
                        n_grabed = self.n_grabed_data % self.Naverage
                    else:
                        n_grabed = self.Naverage
                    self.data_grabed_signal_temp.emit([
                        DataFromPlugins(name=cam_name,
                                        data=[self.data * self.Naverage / n_grabed],
                                        dim=self.data_shape)])

            self.camera_controller.queue_single_buffer(self.buffers[self.current_buffer])

        except Exception as e:
            logger.exception(str(e))

    def update_read_mode(self):
        self.x_axis = self.get_xaxis()
        self.y_axis = self.get_yaxis()

    def setup_image(self):
        binnings = self.camera_controller.AOIBinning.getAvailableValues()
        self.settings.child('camera_settings', 'image_settings', 'binning').setLimits(binnings)
        self.settings.child('camera_settings', 'image_settings', 'bin_x').setLimits(
            (self.camera_controller.AOIHBin.min(), self.camera_controller.AOIHBin.max()))
        self.settings.child('camera_settings', 'image_settings', 'bin_y').setLimits(
            (self.camera_controller.AOIVBin.min(), self.camera_controller.AOIVBin.max()))
        self.settings.child('camera_settings', 'image_settings', 'im_left').setLimits(
            (self.camera_controller.AOILeft.min(), self.camera_controller.AOILeft.max()))
        self.settings.child('camera_settings', 'image_settings', 'im_top').setLimits(
            (self.camera_controller.AOITop.min(), self.camera_controller.AOITop.max()))
        self.settings.child('camera_settings', 'image_settings', 'im_width').setLimits(
            (self.camera_controller.AOIWidth.min(), self.camera_controller.AOIWidth.max()))
        self.settings.child('camera_settings', 'image_settings', 'im_height').setLimits(
            (self.camera_controller.AOIHeight.min(), self.camera_controller.AOIHeight.max()))

    def get_ROI_size_x(self):
        return self.settings.child('camera_settings', 'image_settings', 'im_width').value()

    def get_pixel_size(self):
        """Get the Macro pixel size"""
        return self.camera_controller.PixelWidth.getValue() * self.camera_controller.AOIHBin.getValue()

    def set_image_area(self):

        left = self.settings.child('camera_settings', 'image_settings', 'im_left').value()
        top = self.settings.child('camera_settings', 'image_settings', 'im_top').value()
        width = self.settings.child('camera_settings', 'image_settings', 'im_width').value()
        height = self.settings.child('camera_settings', 'image_settings', 'im_height').value()

        # first set width and height as it modifies the possible top and left
        self.camera_controller.AOIWidth.setValue(width)
        self.camera_controller.AOIHeight.setValue(height)
        self.camera_controller.AOITop.setValue(top)
        self.camera_controller.AOILeft.setValue(left)

        self.settings.child('camera_settings', 'image_settings',
                            'readout_time').setValue(self.camera_controller.ReadoutTime.getValue() * 1000)

    def ini_detector(self, controller=None):
        """
            Initialisation procedure of the detector in four steps :
                * Register callback to get data from camera
                * Get image size and current binning
                * Set and Get temperature from camera
                * Init axes from image

            Returns
            -------
            string list ???
                The initialized status.

            See Also
            --------
            daq_utils.ThreadCommand, hardware1D.DAQ_1DViewer_Picoscope.update_pico_settings
        """
        model = self.settings.child('camera_settings', 'camera_model')
        self.camera_controller = self.ini_detector_init(old_controller=controller,
                                                        new_controller=api.AndorCamera(
                                                            model.opts['limits'].index(model.value())))

        self.ini_camera()

        # %%%%%%% init axes from image
        self.x_axis = self.get_xaxis()
        self.y_axis = self.get_yaxis()

        self.data_grabed_signal_temp.emit([DataFromPlugins(name='Andor SCMOS',
                                                           data=[np.zeros((len(self.y_axis), len(self.x_axis)))],
                                                           dim='Data2D', labels=['dat0'],
                                                           x_axis=self.x_axis,
                                                           y_axis=self.y_axis), ])

        self.emit_status(ThreadCommand('close_splash'))

        info = ""
        initialized = True
        return info, initialized

    def setup_temperature(self):
        enable = self.settings.child('camera_settings', 'temperature_settings', 'enable_cooling').value()
        self.camera_controller.SensorCooling.setValue(enable)
        if not self.temperature_timer.isActive():
            self.temperature_timer.start(2000)  # Timer event fired every 2s
        if enable:
            if self.camera_controller.TemperatureControl.isWritable():
                self.camera_controller.TemperatureControl.setString(self.settings.child('camera_settings', 'temperature_settings', 'set_point').value())
            self.update_temperature()
            # set timer to update temperature info from controller


    def update_temperature(self):
        """
        update temperature status and value. Fired using the temperature_timer every 2s when not grabbing
        """
        temp = self.camera_controller.SensorTemperature.getValue()
        status = self.camera_controller.TemperatureStatus.getString()
        self.settings.child('camera_settings', 'temperature_settings', 'current_value').setValue(temp)
        self.settings.child('camera_settings', 'temperature_settings', 'status').setValue(status)

        self.settings.child('camera_settings',
                            'frame_rate').setValue(self.camera_controller.FrameRate.getValue())


    def ini_camera(self):
        sdk3cam.camReg.regCamera()
        self.camera_controller.init_camera()

        if not self.camera_controller.FullAOIControl.getValue():
            self.settings.child('camera_settings', 'image_settings').opts(readonly=True)

        self.settings.child('camera_settings', 'encoding').setOpts(
            limits=self.camera_controller.SimplePreAmpGainControl.getAvailableValues())
        self.settings.child('camera_settings', 'encoding').setValue(self.camera_controller.SimplePreAmpGainControl.getString())

        # %%%%%% Get image size and current binning
        self.SIZEX, self.SIZEY = self.camera_controller.GetCCDWidth(), self.camera_controller.GetCCDHeight()
        self.settings.child('camera_settings', 'image_settings', 'im_height').setValue(self.SIZEY)
        self.settings.child('camera_settings', 'image_settings', 'im_width').setValue(self.SIZEX)
        self.settings.child('camera_settings', 'image_settings', 'im_left').setValue(1)
        self.settings.child('camera_settings', 'image_settings', 'im_top').setValue(1)
        self.settings.child('camera_settings', 'image_size', 'Nx').setOpts(max=self.SIZEX, default=self.SIZEX, value=self.SIZEX)
        self.settings.child('camera_settings', 'image_size', 'Ny').setOpts(max=self.SIZEY, default=self.SIZEY, value=self.SIZEY)
        #
        # get max exposure range

        self.settings.child('camera_settings', 'exposure').setLimits((self.camera_controller.ExposureTime.min() * 1000,
                                                                      self.camera_controller.ExposureTime.max() * 1000))

        self.setup_image()

        # %%%%%%% Set and Get temperature from camera
        # get temperature range
        statuses = self.camera_controller.TemperatureStatus.getAvailableValues()
        self.settings.child('camera_settings', 'temperature_settings', 'status').setOpts(limits=statuses)
        values = self.camera_controller.TemperatureControl.getAvailableValues()
        self.settings.child('camera_settings', 'temperature_settings', 'set_point').setOpts(limits=values,
                                                                         readonly=self.camera_controller.TemperatureControl.isReadonly())
        self.setup_temperature()

        self.setup_trigger()

        self.setup_shutter()

        self.setup_callback()

    def setup_callback(self):
        if self.callback_thread is not None:
            if self.callback_thread.isRunning():
                self.callback_thread.terminate()

        callback = AndorCallback(self.camera_controller.wait_buffer)
        self.callback_thread = QtCore.QThread()
        callback.moveToThread(self.callback_thread)
        callback.data_sig.connect(
            self.emit_data)  # when the wait for acquisition returns (with data taken), emit_data will be fired

        self.start_waitloop.connect(callback.start)
        self.stop_waitloop.connect(callback.stop)
        self.callback_thread.callback = callback
        self.callback_thread.start()

    def setup_trigger(self):
        self.settings.child('camera_settings', 'trigger',
                            'trigger_mode').setLimits(self.camera_controller.TriggerMode.getAvailableValues())
        self.settings.child('camera_settings', 'trigger',
                            'ext_trigger_delay').setLimits((self.camera_controller.ExternalTriggerDelay.min(),
                                                            self.camera_controller.ExternalTriggerDelay.max()))
        self.set_trigger()

    def setup_shutter(self):
        modes = self.camera_controller.ShutterMode.getAvailableValues()
        output_modes = self.camera_controller.ShutterOutputMode.getAvailableValues()
        self.settings.child('camera_settings', 'shutter', 'shutter_mode').setOpts(limits=modes)
        self.settings.child('camera_settings', 'shutter', 'shutter_on_ext_trigger').setOpts(limits=output_modes)
        self.set_shutter()

    def set_shutter(self):
        self.camera_controller.ShutterMode.setString(self.settings.child('camera_settings', 'shutter', 'shutter_mode').value())
        self.camera_controller.ShutterOutputMode.setString(self.settings.child('camera_settings', 'shutter', 'shutter_on_ext_trigger').value())

    def close(self):
        """

        """
        self.temperature_timer.stop()
        QtWidgets.QApplication.processEvents()
        self.camera_controller.close()

    def get_xaxis(self):
        """
            Obtain the horizontal axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the horizontal camera pixels.
        """
        if self.camera_controller is not None:
            # if self.control_type == "camera":
            Nx = self.settings.child('camera_settings', 'image_size', 'Nx').value()
            self.x_axis = Axis(data=np.linspace(0, Nx - 1, Nx, dtype=np.int), label='Pixels')

            self.emit_x_axis()
        else:
            raise (Exception('controller not defined'))
        return self.x_axis

    def get_yaxis(self):
        """
            Obtain the vertical axis of the image.

            Returns
            -------
            1D numpy array
                Contains a vector of integer corresponding to the vertical camera pixels.
        """
        if self.camera_controller is not None:
            Ny = self.settings.child('camera_settings', 'image_size', 'Ny').value()
            self.y_axis = Axis(data=np.linspace(0, Ny - 1, Ny, dtype=np.int), label='Pixels')
            self.emit_y_axis()
        else:
            raise (Exception('Camera not defined'))
        return self.y_axis


    def free_buffers(self):
        # empty buffers if that was not the case
        while len(self.buffers) != 0:
            self.buffers.pop(0)


    def prepare_data(self):
        sizex = self.settings.child('camera_settings', 'image_settings', 'im_width').value()
        sizey = self.settings.child('camera_settings', 'image_settings', 'im_height').value()
        if sizex != self.camera_controller.AOIWidth.getValue():
            logger.error("Memory size doesn't correspond to image size")
            return False
        if sizey != self.camera_controller.AOIHeight.getValue():
            logger.error("Memory size doesn't correspond to image size")
            return False

        # %% Initialize data: self.data for the memory to store new data and self.data_average to store the average data
        image_size = sizex * sizey
        self.data = np.zeros((sizey, sizex), dtype=np.float)

        bufSize = self.camera_controller.ImageSizeBytes.getValue()
        if not self.buffers or self.buffers[0].size != bufSize or self._reset_buffers_cmd:
            if not not self.buffers:
                self.free_buffers()
                self.camera_controller.flush()
                self._reset_buffers_cmd = False
                self.settings.child('camera_settings', 'reset_buffers').setValue(False)

            self.current_buffer = -1
            self.buffers = []
            self.buffers_pointer = []
            self._Nbuffers = self.settings.child('camera_settings', 'buffer_number').value()
            for ind in range(self._Nbuffers):
                self.buffers.append(zeros_aligned(bufSize, 8, 'uint8'))
                self.buffers_pointer.append(self.buffers[-1].ctypes.data)
                self.camera_controller.queue_single_buffer(self.buffers[-1])

        data_shape = 'Data2D' if sizey != 1 else 'Data1D'
        if data_shape != self.data_shape:
            self.data_shape = data_shape
            # init the viewers
            self.data_grabed_signal_temp.emit([DataFromPlugins(
                name=self.settings.child('camera_settings', 'camera_model').value(),
                data=[np.squeeze(self.data.reshape((sizex, sizey)).astype(np.float))], dim=self.data_shape)])

        return True

    def grab_data(self, Naverage=1, **kwargs):
        """
            Start new acquisition in two steps :
                * Initialize data: self.data for the memory to store new data and self.data_average to store the average data
                * Start acquisition with the given exposure in ms, in "1d" or "2d" mode

            =============== =========== =============================
            **Parameters**   **Type**    **Description**
            Naverage         int         Number of images to average
            =============== =========== =============================

            See Also
            --------
            daq_utils.ThreadCommand
        """
        try:
            self.camera_done = False
            self.Naverage = Naverage
            if 'live' in kwargs:
                self.live = kwargs['live']
            if 'wait_time' in kwargs:
                self.wait_time = kwargs['wait_time']


            self.n_grabed_data = 0
            self.n_grabed_frame_rate = 0
            self.start_time = perf_counter()  # to check if one should refresh output
            self.temperature_timer.stop()

            if self.camera_controller.CameraAcquiring.getValue():
                self.camera_controller.AcquisitionStop()

            self.set_image_area()
            status = self.prepare_data()
            if not status:
                return

            self.camera_controller.CycleMode.setString('Continuous')

            self.camera_controller.AcquisitionStart()
            if self.live:
                self.start_waitloop.emit(-1, self.wait_time)  # will trigger the waitfor acquisition
            else:
                self.start_waitloop.emit(Naverage, 0)  # will trigger the waitfor acquisition
        except Exception as e:
            self.emit_status(ThreadCommand('Update_Status', [str(e), "log"]))

    def stop(self):
        """
            stop the camera's actions.
        """
        try:
            self.stop_waitloop.emit()
            if self.camera_controller.CameraAcquiring.getValue():
                self.camera_controller.AcquisitionStop()
            QtWidgets.QApplication.processEvents()
            self.temperature_timer.start(2000)

        except:
            pass
        return ""


class AndorCallback(QtCore.QObject):
    """

    """
    data_sig = QtCore.Signal(list)

    def __init__(self, wait_fn):
        super().__init__()
        self.wait_fn = wait_fn
        self.running = False

    def start(self, naverage, wait_time=0):
        self.running = True

        self.wait_for_acquisition(naverage, wait_time)

    def stop(self):
        self.running = False

    def wait_for_acquisition(self, naverage, wait_time):
        ind_grab = 0
        while True:
            if naverage == -1:  # continuous grab
                if not self.running:
                    break
            else:
                if ind_grab >= naverage or not self.running:
                    break
            pData = self.wait_fn()
            if not not pData:
                ind_grab += 1
                #print(f'ind_grab_thread:{ind_grab}')
                self.data_sig.emit([pData])
                QtCore.QThread.msleep(wait_time)




#!/usr/bin/python

###############
# sdk3cam.py
#
# Copyright David Baddeley, 2012
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################

from __future__ import absolute_import
import logging
from . import sdk3


class ATProperty(object):
    def connect(self, handle, propertyName):
        self.handle = handle
        self.propertyName = propertyName
        
    def isImplemented(self):
        return bool(sdk3.IsImplemented(self.handle, self.propertyName).value)

    def isWritable(self):
        return bool(sdk3.IsWritable(self.handle, self.propertyName).value)

    def isReadable(self):
        return bool(sdk3.IsReadable(self.handle, self.propertyName).value)

    def isReadonly(self):
        return bool(sdk3.IsReadOnly(self.handle, self.propertyName).value)

        


class ATInt(ATProperty):
    def getValue(self):
        return sdk3.GetInt(self.handle, self.propertyName).value

    def setValue(self, val):
        sdk3.SetInt(self.handle, self.propertyName, val)

    def max(self):
        return sdk3.GetIntMax(self.handle, self.propertyName).value

    def min(self):
        return sdk3.GetIntMin(self.handle, self.propertyName).value


class ATBool(ATProperty):
    def getValue(self):
        return sdk3.GetBool(self.handle, self.propertyName).value > 0

    def setValue(self, val):
        sdk3.SetBool(self.handle, self.propertyName, val)


class ATFloat(ATProperty):
    def getValue(self):
        return sdk3.GetFloat(self.handle, self.propertyName).value

    def setValue(self, val):
        sdk3.SetFloat(self.handle, self.propertyName, val)

    def max(self):
        return sdk3.GetFloatMax(self.handle, self.propertyName).value

    def min(self):
        return sdk3.GetFloatMin(self.handle, self.propertyName).value


class ATString(ATProperty):
    def getValue(self):
        return sdk3.GetString(self.handle, self.propertyName, 255).value

    def setValue(self, val):
        sdk3.SetString(self.handle, self.propertyName, val)

    def maxLength(self):
        return sdk3.GetStingMaxLength(self.handle, self.propertyName).value


class ATEnum(ATProperty):
    def getIndex(self):
        return sdk3.GetEnumIndex(self.handle, self.propertyName).value

    def setIndex(self, val):
        sdk3.SetEnumIndex(self.handle, self.propertyName, val)

    def getString(self):
        return self.__getitem__(self.getIndex())

    def setString(self, val):
        sdk3.SetEnumString(self.handle, self.propertyName, val)

    def __len__(self):
        return sdk3.GetEnumCount(self.handle, self.propertyName).value

    def __getitem__(self, key):
        return sdk3.GetEnumStringByIndex(self.handle, self.propertyName, key, 255).value

    def getAvailableValues(self):
        n = sdk3.GetEnumCount(self.handle, self.propertyName).value

        return [sdk3.GetEnumStringByIndex(self.handle, self.propertyName, i, 255).value for i in range(n) if sdk3.IsEnumIndexAvailable(self.handle, self.propertyName, i).value]


class ATCommand(ATProperty):
    def __call__(self):
        return sdk3.Command(self.handle, self.propertyName)


class camReg(object):
    # keep track of the number of cameras initialised so we can initialise and finalise the library
    numCameras = 0

    @classmethod
    def regCamera(cls):
        if cls.numCameras == 0:
            sdk3.InitialiseLibrary()

        cls.numCameras += 1

    @classmethod
    def unregCamera(cls):
        cls.numCameras -= 1
        if cls.numCameras == 0:
            sdk3.FinaliseLibrary()

camReg.regCamera()

def getNumCameras():
    return sdk3.GetInt(sdk3.AT_HANDLE_SYSTEM, 'DeviceCount').value


def getSoftwareVersion():
    return sdk3.GetString(sdk3.AT_HANDLE_SYSTEM, 'SoftwareVersion', 255).value


class SDK3Camera(object):
    def __init__(self, camNum):
        '''camera initialisation - note that this should be called  from derived classes
        *AFTER* the properties have been defined'''
        # camReg.regCamera() #initialise the library if needed
        self.camNum = camNum

    def init_camera(self):
        self.handle = sdk3.Open(self.camNum)
        self.connectProperties()

    def connectProperties(self):
        not_implemented = []
        for name, var in self.__dict__.items():
            if isinstance(var, ATProperty):
                var.connect(self.handle, name)

    def shutdown(self):
        sdk3.Close(self.handle)
        # camReg.unregCamera()

    def close(self):
        self.shutdown()
        camReg.unregCamera()

    # def register_callback(self, callback):
    #     sdk3.Re






#!/usr/bin/python

##################
# AndorCam.py
#
# Copyright David Baddeley, 2009
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##################
from ctypes import POINTER, c_int, c_uint, c_double
"""
Note that you need atcore and atutility. 
You obtain these files from MyAndor under Software > Drivers and 3rd Party > Andor DriverPack 3 (sCMOS)

Next add this directory to your system PATH
In Windows this is accomplished through
Control Panel / System / Advanced System Settings / Environment Variables / System Variables
and add to path: 
C:\Program Files\Andor Driver Pack 3
"""
import sys
import ctypes
import platform
from pathlib import Path
import os
_stdcall_libraries = {}

arch, plat = platform.architecture()

sdk_path = 'C:\\Program Files\\Andor SDK3'

if plat.startswith('Windows'):
    try:
        if sdk_path not in os.environ['path']:
            os.environ['path'] += ';' + sdk_path
        _stdcall_libraries['ATCORE'] = ctypes.windll.LoadLibrary('atcore.dll')
        _stdcall_libraries['ATUTIL'] = ctypes.windll.LoadLibrary('atutility.dll')
    except OSError as e:
        raise OSError(f'Andor SCMOS atcore and atutlity DLL not found: {str(e)}')

else:
    try:
        _stdcall_libraries['ATCORE'] = ctypes.CDLL('atcore.so')
        _stdcall_libraries['ATUTIL'] = ctypes.CDLL('atutility.so')
    except OSError as e:
        raise OSError(f'Andor SCMOS atcore and atutlity DLL not found: {str(e)}')


# typedefs
AT_H = ctypes.c_int
AT_BOOL = ctypes.c_int
AT_64 = ctypes.c_int64
AT_U8 = ctypes.c_uint8
AT_WC = ctypes.c_wchar


# Defines
errorCodes = {}


def errCode(name, value):
    errorCodes[value] = name


AT_INFINITE = 0xFFFFFFFF
AT_CALLBACK_SUCCESS = 0

AT_TRUE = 1
AT_FALSE = 0

AT_SUCCESS = 0
errCode('AT_ERR_NOTINITIALISED', 1)
errCode('AT_ERR_NOTIMPLEMENTED', 2)
errCode('AT_ERR_READONLY', 3)
errCode('AT_ERR_NOTREADABLE', 4)
errCode('AT_ERR_NOTWRITABLE', 5)
errCode('AT_ERR_OUTOFRANGE', 6)
errCode('AT_ERR_INDEXNOTAVAILABLE', 7)
errCode('AT_ERR_INDEXNOTIMPLEMENTED', 8)
errCode('AT_ERR_EXCEEDEDMAXSTRINGLENGTH', 9)
errCode('AT_ERR_CONNECTION', 10)
errCode('AT_ERR_NODATA', 11)
errCode('AT_ERR_INVALIDHANDLE', 12)
errCode('AT_ERR_TIMEDOUT', 13)
errCode('AT_ERR_BUFFERFULL', 14)
errCode('AT_ERR_INVALIDSIZE', 15)
errCode('AT_ERR_INVALIDALIGNMENT', 16)
errCode('AT_ERR_COMM', 17)
errCode('AT_ERR_STRINGNOTAVAILABLE', 18)
errCode('AT_ERR_STRINGNOTIMPLEMENTED', 19)

errCode('AT_ERR_NULL_FEATURE', 20)
errCode('AT_ERR_NULL_HANDLE', 21)
errCode('AT_ERR_NULL_IMPLEMENTED_VAR', 22)
errCode('AT_ERR_NULL_READABLE_VAR', 23)
errCode('AT_ERR_NULL_READONLY_VAR', 24)
errCode('AT_ERR_NULL_WRITABLE_VAR', 25)
errCode('AT_ERR_NULL_MINVALUE', 26)
errCode('AT_ERR_NULL_MAXVALUE', 27)
errCode('AT_ERR_NULL_VALUE', 28)
errCode('AT_ERR_NULL_STRING', 29)
errCode('AT_ERR_NULL_COUNT_VAR', 30)
errCode('AT_ERR_NULL_ISAVAILABLE_VAR', 31)
errCode('AT_ERR_NULL_MAXSTRINGLENGTH', 32)
errCode('AT_ERR_NULL_EVCALLBACK', 33)
errCode('AT_ERR_NULL_QUEUE_PTR', 34)
errCode('AT_ERR_NULL_WAIT_PTR', 35)
errCode('AT_ERR_NULL_PTRSIZE', 36)
errCode('AT_ERR_NOMEMORY', 37)

errCode('AT_ERR_HARDWARE_OVERFLOW', 100)


class CameraError(Exception):
    def __init__(self, fcnName, errNo):
        self.errNo = errNo
        self.fcnName = fcnName

    def __str__(self):
        return 'when calling %s - %s' % (self.fcnName, errorCodes[self.errNo])


# special case for buffer timeout
AT_ERR_TIMEDOUT = 13
AT_ERR_NODATA = 11


class TimeoutError(CameraError):
    pass


AT_HANDLE_UNINITIALISED = -1
AT_HANDLE_SYSTEM = 1

### Functions ###
STRING = POINTER(AT_WC)

# classes so that we do some magic and automatically add byrefs etc ... can classify outputs


class _meta(object):
    pass


class OUTPUT(_meta):
    def __init__(self, val):
        self.type = val
        self.val = POINTER(val)

    def getVar(self, bufLen=0):
        v = self.type()
        return v, ctypes.byref(v)


class _OUTSTRING(OUTPUT):
    def __init__(self):
        self.val = STRING

    def getVar(self, bufLen):
        v = ctypes.create_unicode_buffer(bufLen)
        return v, v


OUTSTRING = _OUTSTRING()


class _OUTSTRLEN(_meta):
    def __init__(self):
        self.val = c_int


OUTSTRLEN = _OUTSTRLEN()


def stripMeta(val):
    if isinstance(val, _meta):
        return val.val
    else:
        return val


class dllFunction(object):
    def __init__(self, name, args=[], argnames=[], lib='ATCORE'):
        self.f = getattr(_stdcall_libraries[lib], name)
        self.f.restype = c_int
        self.f.argtypes = [stripMeta(a) for a in args]

        self.fargs = args
        self.fargnames = argnames
        self.name = name

        self.inp = [not isinstance(a, OUTPUT) for a in args]
        self.in_args = [a for a in args if not isinstance(a, OUTPUT)]
        self.out_args = [a for a in args if isinstance(a, OUTPUT)]

        self.buf_size_arg_pos = -1
        for i in range(len(self.in_args)):
            if isinstance(self.in_args[i], _OUTSTRLEN):
                self.buf_size_arg_pos = i

        ds = name + '\n\nArguments:\n===========\n'
        for i in range(len(args)):
            an = ''
            if i < len(argnames):
                an = argnames[i]
            ds += '\t%s\t%s\n' % (args[i], an)

        self.f.__doc__ = ds

    def __call__(self, *args):
        ars = []
        i = 0
        ret = []

        if self.buf_size_arg_pos >= 0:
            bs = args[self.buf_size_arg_pos]
        else:
            bs = 255

        for j in range(len(self.inp)):
            if self.inp[j]:  # an input
                ars.append(args[i])
                i += 1
            else:  # an output
                r, ar = self.fargs[j].getVar(bs)
                ars.append(ar)
                ret.append(r)
                # print r, r._type_

        # print ars
        res = self.f(*ars)
        # print res

        if not res == AT_SUCCESS:
            if res == AT_ERR_TIMEDOUT or res == AT_ERR_NODATA:
                # handle timeouts as a special case, as we expect to get them
                raise TimeoutError(self.name, res)
            else:
                raise CameraError(self.name, res)

        if len(ret) == 0:
            return None
        if len(ret) == 1:
            return ret[0]
        else:
            return ret


def dllFunc(name, args=[], argnames=[], lib='ATCORE'):
    f = dllFunction(name, args, argnames, lib)
    globals()[name[3:]] = f


dllFunc('AT_InitialiseLibrary')
dllFunc('AT_FinaliseLibrary')

dllFunc('AT_Open', [c_int, OUTPUT(AT_H)])
dllFunc('AT_Close', [AT_H])

dllFunc('AT_IsImplemented', [AT_H, STRING, OUTPUT(AT_BOOL)])
dllFunc('AT_IsReadable', [AT_H, STRING, OUTPUT(AT_BOOL)])
dllFunc('AT_IsWritable', [AT_H, STRING, OUTPUT(AT_BOOL)])
dllFunc('AT_IsReadOnly', [AT_H, STRING, OUTPUT(AT_BOOL)])

dllFunc('AT_SetInt', [AT_H, STRING, AT_64])
dllFunc('AT_GetInt', [AT_H, STRING, OUTPUT(AT_64)])
dllFunc('AT_GetIntMax', [AT_H, STRING, OUTPUT(AT_64)])
dllFunc('AT_GetIntMin', [AT_H, STRING, OUTPUT(AT_64)])

dllFunc('AT_SetFloat', [AT_H, STRING, c_double])
dllFunc('AT_GetFloat', [AT_H, STRING, OUTPUT(c_double)])
dllFunc('AT_GetFloatMax', [AT_H, STRING, OUTPUT(c_double)])
dllFunc('AT_GetFloatMin', [AT_H, STRING, OUTPUT(c_double)])

dllFunc('AT_SetBool', [AT_H, STRING, AT_BOOL])
dllFunc('AT_GetBool', [AT_H, STRING, OUTPUT(AT_BOOL)])

dllFunc('AT_SetEnumerated', [AT_H, STRING, c_int])
dllFunc('AT_SetEnumeratedString', [AT_H, STRING, STRING])
dllFunc('AT_GetEnumerated', [AT_H, STRING, OUTPUT(c_int)])
dllFunc('AT_GetEnumeratedCount', [AT_H, STRING, OUTPUT(c_int)])
dllFunc('AT_IsEnumeratedIndexAvailable', [AT_H, STRING, c_int, OUTPUT(AT_BOOL)])
dllFunc('AT_IsEnumeratedIndexImplemented', [AT_H, STRING, c_int, OUTPUT(AT_BOOL)])
dllFunc('AT_GetEnumeratedString', [AT_H, STRING, c_int, OUTSTRING, OUTSTRLEN])

dllFunc('AT_SetEnumIndex', [AT_H, STRING, c_int])
dllFunc('AT_SetEnumString', [AT_H, STRING, STRING])
dllFunc('AT_GetEnumIndex', [AT_H, STRING, OUTPUT(c_int)])
dllFunc('AT_GetEnumCount', [AT_H, STRING, OUTPUT(c_int)])
dllFunc('AT_IsEnumIndexAvailable', [AT_H, STRING, c_int, OUTPUT(AT_BOOL)])
dllFunc('AT_IsEnumIndexImplemented', [AT_H, STRING, c_int, OUTPUT(AT_BOOL)])
dllFunc('AT_GetEnumStringByIndex', [AT_H, STRING, c_int, OUTSTRING, OUTSTRLEN])

dllFunc('AT_Command', [AT_H, POINTER(AT_WC)])

dllFunc('AT_SetString', [AT_H, STRING, STRING])
dllFunc('AT_GetString', [AT_H, STRING, OUTSTRING, OUTSTRLEN])
dllFunc('AT_GetStringMaxLength', [AT_H, STRING, OUTPUT(c_int)])

dllFunc('AT_QueueBuffer', [AT_H, POINTER(AT_U8), c_int])
dllFunc('AT_WaitBuffer', [AT_H, OUTPUT(POINTER(AT_U8)), OUTPUT(c_int), c_uint])
dllFunc('AT_Flush', [AT_H])

#####################################
# Utility library (for unpacking etc ...)
dllFunc('AT_InitialiseUtilityLibrary', lib='ATUTIL')
dllFunc('AT_FinaliseUtilityLibrary', lib='ATUTIL')
dllFunc('AT_ConvertBuffer', [POINTER(AT_U8), POINTER(AT_U8), AT_64, AT_64, AT_64, STRING, STRING], lib='ATUTIL')
dllFunc('AT_ConvertBufferUsingMetadata', [POINTER(AT_U8), POINTER(AT_U8), AT_64, STRING], lib='ATUTIL')

# Initialize the utility library
InitialiseUtilityLibrary()



#!/usr/bin/python

###############
# testNeo.py
#
# Copyright David Baddeley, 2012
# d.baddeley@auckland.ac.nz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################

from __future__ import absolute_import
import time
from copy import copy
import numpy as np
import logging
logging.basicConfig(format='%(asctime)s.%(msecs)03d %(filename)s/%(funcName)s:%(lineno)d %(message)s', datefmt='%Y-%m-%d %H:%M:%S',
                    level=logging.WARNING)


def basicNeoTest(nframe, nbuffer):
    logging.info('Importing Camera')
    from pymodaq_plugins_andor.hardware.andor_sdk3 import api as andor
    from pymodaq_plugins_andor.hardware.andor_sdk3 import sdk3cam
    # %%
    logging.info('Initialising Camera')
    cam = andor.AndorCamera(0, nbuffer)
    sdk3cam.camReg.regCamera()
    cam.init_camera()


    cam.SetIntegTime(1)  # exposure milliseconds
    # cam.PixelReadoutRate.setIndex(2)
    # %%
    logging.info('Starting Exposure')
    cam.StartExposure()

    # this is one single frame, no more
    rbuf, cbuf = cam.GetPicWidth(), cam.GetPicHeight()
    buf = np.empty((rbuf, cbuf), np.uint16)

    # output image array
    imgs = []  # np.empty((nbuffer,cbuf,rbuf),np.uint16)

    logging.info('Starting Extraction loop')
    time.sleep(1)  # will get blank images without at least 62.5ms delay! (60.0ms doesn't work)
    for i in range(nframe):
        while cam.ExpReady():
            cam.ExtractColor(buf, 1)
            # WARNING: copy(buf) is absolutely necessary or all images will be the same!
            imgs.append(copy(buf).reshape((cbuf, rbuf), order="C"))

    cam.shutdown()
    time.sleep(.05)
    sdk3cam.camReg.unregCamera()

    return imgs


if __name__ == '__main__':
    import h5py
    from matplotlib.pyplot import figure, draw, pause
    import cv2
    imgs = basicNeoTest(3, 1)
    print('{} images returned'.format(len(imgs)))

    with h5py.File('blah.h5', 'w', libver='latest') as f:
        f['/imgs'] = imgs

    if 0:
        fg = figure()
        ax = fg.gca()
        hi = ax.imshow(imgs[0], cmap='gray')
        for i, I in enumerate(imgs):
            hi.set_data(I)
            # fg.colorbar(hi)
            ax.set_title(i)
            draw()
            pause(0.01)

    if True:
        for I in imgs:
            cv2.imshow('frame', I)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break



# -*- coding: utf-8 -*-
#   AndoriDus - A Python wrapper for Andor's scientific cameras
#
#   Original code by
#   Copyright (C) 2009  Hamid Ohadi
#
#   Adapted for iDus, qtlab and Windows XP
#   2010 Martijn Schaafsma
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#   Adapted for PyMoDAQ
#   2020 Sébastien Weber


# Modules for Andor functionality
import ctypes
from ctypes import windll, c_int, c_char, byref, c_long, \
    pointer, c_float, c_char_p, cdll
import sys
import time
from pathlib import Path
import platform
import os

"""
NOTE:

If controlling the shamrock through i2c it is important that both
the camera and spectrograph are being controlled through the same 
calling program and that the DLLs used are contained in the same 
working folder.

The camera MUST be initialized before attempting to communicate 
with the Shamrock.

"""
try:
    # Check operating system and load library
    if platform.system() == "Linux":
        dllpath = Path("/usr/local/lib/libandor.so")
        _dll = cdll.LoadLibrary(str(dllpath))
    elif platform.system() == "Windows":
        dllpath = Path('C:\\Program Files\\Andor SDK')
        if platform.machine() == "AMD64":
            dllname = "atmcd64d.dll"
        else:
            dllname = "atmcd32d.dll"
        dllpath = dllpath.joinpath(dllname)
        _dll = windll.LoadLibrary(str(dllpath))

except Exception as e:
    raise(f'Could not import Andor CDD library: {str(e)}')
            

# class AndorIdus(Instrument):
class AndorSDK:
    """
    Andor class which is meant to provide the Python version of the same
    functions that are defined in the Andor's SDK. Extensive documentation
    on the functions used and error codes can be
    found in the Andor SDK Users Guide
    """

    def __init__(self):
        """
        """
    @classmethod
    def init_camera(cls):
        # Initialize the device
        tekst = c_char()
        print("Initializing iDus...", )
        error = _dll.Initialize(byref(tekst))
        if error != 20002:
            raise IOError(ERROR_CODE[error])


    def __del__(self):
        _dll.ShutDown()

    def close(self):
        error = _dll.ShutDown()

    # Manage camera

    @classmethod
    def GetAvailableCameras(cls):
        Ncam = c_long()
        error = _dll.GetAvailableCameras(byref(Ncam))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return Ncam.value

    @classmethod
    def GetCameraHandle(cls, index):
        handle = c_long()
        index = c_long(index)
        error = _dll.GetCameraHandle(index, byref(handle))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return handle.value

    @classmethod
    def SetCurrentCamera(cls, handle):
        handle = c_long(handle)
        error = _dll.SetCurrentCamera(handle)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    @classmethod
    def GetCurrentCamera(cls):
        handle = c_long()
        error = _dll.GetCurrentCamera(byref(handle))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return handle.value

    @classmethod
    def GetCamerasInfo(cls):
        cameralist = []
        for ind in range(cls.GetAvailableCameras()):
            handle = cls.GetCameraHandle(ind)
            cls.SetCurrentCamera(handle)
            cls.init_camera()
            cameralist.append(dict(handle=handle,
                                   serial=cls.GetCameraSerialNumber(),
                                   model=cls.GetHeadModel()))
        return cameralist

    # Get Camera properties
    @classmethod
    def GetCameraSerialNumber(cls):
        '''
        Returns the serial number of the camera

        Input:
            None

        Output:
            (int) : Serial number of the camera
        '''
        serial = c_int()
        error = _dll.GetCameraSerialNumber(byref(serial))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return serial.value

    @classmethod
    def GetHeadModel(cls):
        '''
        Returns the camera head model name

        Input:
            None

        Output:
            (str) : model name
        '''
        model = ctypes.create_string_buffer(128)
        error = _dll.GetHeadModel(byref(model))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return model.value.decode()

    def GetDetector(self):
        '''
        Returns the camera sensor size

        Input:
            None

        Output:
            (int,int) : x and y size (hor and ver)
        '''
        xpixels = c_int()
        ypixels = c_int()
        error = _dll.GetDetector(byref(xpixels), byref(ypixels))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

        return (xpixels.value, ypixels.value)

    def GetMaximumBinning(self, readmode=0, horver=1):
        '''
        Returns the max binning size in horizontal (horver=0) or vertical (horver=1) given the readout mode

        Input:
            int: readmode (see SetReadMode)
            int: 0 to retrieve horizontal binning limit, 1 to retreive limit in the vertical.

        Output:
            (str,int) : error and binning value
        '''
        maxbinning = c_int()
        error = _dll.GetMaximumBinning(readmode, horver, byref(maxbinning))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return (ERROR_CODE[error], maxbinning.value)

    def GetNumberHSSpeeds(self):
        '''
        Returns the number of HS speeds

        Input:
            None

        Output:
            (int) : the number of HS speeds
        '''
        noHSSpeeds = c_int()
        error = _dll.GetNumberHSSpeeds(self._channel, self._outamp,
                                            byref(noHSSpeeds))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return noHSSpeeds.value

    def GetNumberVSSpeeds(self):
        '''
        Returns the number of VS speeds

        Input:
            None

        Output:
            (int) : the number of VS speeds
        '''
        noVSSpeeds = c_int()
        error = _dll.GetNumberVSSpeeds(byref(noVSSpeeds))
        self._noVSSpeeds = noVSSpeeds.value
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return noVSSpeeds.value

    # Cooler and temperature
    def CoolerON(self):
        '''
        Switches the cooler on

        Input:
            None

        Output:
            None
        '''
        error = _dll.CoolerON()
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def CoolerOFF(self):
        '''
        Switches the cooler off

        Input:
            None

        Output:
            None
        '''
        error = _dll.CoolerOFF()
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetCoolerMode(self, mode):
        '''
        Set the cooler mode

        Input:
            mode (int) : cooler modus

        Output:
            None
        '''
        error = _dll.SetCoolerMode(mode)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def IsCoolerOn(self):
        '''
        Returns cooler status

        Input:
            None

        Output:
            (int) : Cooler status
        '''
        iCoolerStatus = c_int()
        error = _dll.IsCoolerOn(byref(iCoolerStatus))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return iCoolerStatus.value

    def GetTemperatureRange(self):
        '''
        Returns the temperature range in degrees Celcius

        Input:
            None

        Output:
            (int,int) : temperature min and max in degrees Celcius
        '''
        ctemperature_min = c_int()
        ctemperature_max = c_int()

        error = _dll.GetTemperatureRange(byref(ctemperature_min), byref(ctemperature_max))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], (ctemperature_min.value, ctemperature_max.value)

    def GetTemperature(self):
        '''
        Returns the temperature in degrees Celcius

        Input:
            None

        Output:
            (int) : temperature in degrees Celcius
        '''
        ctemperature = c_int()
        error = _dll.GetTemperature(byref(ctemperature))
        return ERROR_CODE[error], ctemperature.value

    def SetTemperature(self, temperature):
        '''
        Set the working temperature of the camera

        Input:
            temparature (int) : temperature in degrees Celcius

        Output:
            None
        '''
        #        ctemperature = c_int(temperature)
        error = _dll.SetTemperature(int(temperature))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    ###### Single Parameters Set ######
    def SetAccumulationCycleTime(self, time_):
        '''
        Set the accumulation cycle time

        Input:
            time_ (float) : the accumulation cycle time in seconds

        Output:
            None
        '''
        error = _dll.SetAccumulationCycleTime(c_float(time_))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetAcquisitionMode(self, mode):
        '''
        Set the acquisition mode of the camera

        Input:
            mode (int) : acquisition mode

        Output:
            None
        '''
        error = _dll.SetAcquisitionMode(int(mode))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetDriverEvent(self, hevent):
        '''
        passes a Win32 Event handle to the SDK

        Input:
            hevent (win32 event) : event handle (created with win32event.CreateEvent

        Output:
            None
        '''
        error = _dll.SetDriverEvent(hevent)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def WaitForAcquisition(self):
        '''
        wait for acquisition to return (to be called from another thread)

        Output:
            error (str): error as a string
        '''
        error = _dll.WaitForAcquisition()
        return ERROR_CODE[error]

    def CancelWait(self):
        '''
        cancel the wait for acquisition to return

        Output:
            error (str): error as a string
        '''
        error = _dll.CancelWait()
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def SetADChannel(self, index):
        '''
        Set the A-D channel for acquisition

        Input:
            index (int) : AD channel

        Output:
            None
        '''
        error = _dll.SetADChannel(index)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetEMAdvanced(self, gainAdvanced):
        '''
        Enable/disable access to the advanced EM gain levels

        Input:
            gainAdvanced (int) : 1 or 0 for true or false

        Output:
            None
        '''
        error = _dll.SetEMAdvanced(gainAdvanced)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetEMCCDGainMode(self, gainMode):
        '''
        Set the gain mode

        Input:
            gainMode (int) : mode

        Output:
            None
        '''
        error = _dll.SetEMCCDGainMode(gainMode)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetExposureTime(self, time_):
        '''
        Set the exposure time in seconds

        Input:
            time_ (float) : The exposure time in seconds

        Output:
            None
        '''
        error = _dll.SetExposureTime(c_float(time_))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def GetMaximumExposure(self):
        '''
        Get the maximum settable exposure time in seconds


        Output:
            str   : error code as string
            float : Will contain the Maximum exposure value on return.
        '''
        maxexpo = c_float()
        error = _dll.GetMaximumExposure(byref(maxexpo))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], maxexpo.value

    def SetFrameTransferMode(self, frameTransfer):
        '''
        Enable/disable the frame transfer mode

        Input:
            frameTransfer (int) : 1 or 0 for true or false

        Output:
            None
        '''
        error = _dll.SetFrameTransferMode(frameTransfer)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetImageRotate(self, iRotate):
        '''
        Set the modus for image rotation

        Input:
            iRotate (int) : 0 for no rotation, 1 for 90 deg cw, 2 for 90 deg ccw

        Output:
            None
        '''
        error = _dll.SetImageRotate(iRotate)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetKineticCycleTime(self, time_):
        '''
        Set the Kinetic cycle time in seconds

        Input:
            time_ (float) : The cycle time in seconds

        Output:
            None
        '''
        error = _dll.SetKineticCycleTime(c_float(time_))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetNumberAccumulations(self, number):
        '''
        Set the number of scans accumulated in memory,
        for kinetic and accumulate modes

        Input:
            number (int) : The number of accumulations

        Output:
            None
        '''
        error = _dll.SetNumberAccumulations(int(number))
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetNumberKinetics(self, numKin):
        '''
        Set the number of scans accumulated in memory for kinetic mode

        Input:
            number (int) : The number of accumulations

        Output:
            None
        '''
        error = _dll.SetNumberKinetics(numKin)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetOutputAmplifier(self, index):
        '''
        Specify which amplifier to use if EMCCD is enabled

        Input:
            index (int) : 0 for EMCCD, 1 for conventional

        Output:
            None
        '''
        error = _dll.SetOutputAmplifier(index)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetReadMode(self, mode):
        '''
        Set the read mode of the camera

        Input:
            mode (int) : 0 Full Vertical Binning
                         1 Multi-Track
                         2 Random-track
                         3 Single-Track
                         4 Image

        Output:
            error
        '''
        error = _dll.SetReadMode(mode)
        self._ReadMode = mode
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def SetSingleTrack(self, center, height=10):
        '''
        Set the single track mode of the camera

        Input:
            center (int) : center pixel of the track
            height (int) : height of the track

        Output:
            error
        '''
        error = _dll.SetSingleTrack(int(center), int(height))

        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def SetMultiTrack(self, N, height, offset):
        '''
        Set the multi track mode of the camera

        Input:
            N (int) : number of tracks
            height (int) : height of each track
            offset (int) : vertical offset

        Output:
            error
            bottom (int): first pixel of first row
            gap (int) : number of rows between each track (could be 0)
        '''
        bottom = c_int()
        gap = c_int()

        error = _dll.SetMultiTrack(int(N), int(height), int(offset),
                                        byref(bottom), byref(gap))

        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error], bottom.value, gap.value

    def SetImage(self, binx, biny, startx, endx, starty, endy):
        '''
        Set the image mode of the camera

        Input:
            binx (int) : binning along x
            biny (int) : binning along y
            startx (int) : left area position
            endx (int) : right area position
            starty (int) : bottom area position
            endy (int) : top area position
            
        Output:
            error
        '''
        error = _dll.SetImage(int(binx), int(biny), int(startx),
                                   int(endx), int(starty), int(endy))

        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def SetTriggerMode(self, mode):
        '''
        Set the trigger mode

        Input:
            mode (int) : 0 Internal
                         1 External
                         2 External Start (only in Fast Kinetics mode)

        Output:
            None
        '''
        error = _dll.SetTriggerMode(mode)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    ###### Single Parameters Get ######

    def GetAccumulationProgress(self):
        '''
        Returns the number of completed accumulations

        Input:
            None

        Output:
            (int) : The number of accumulations
        '''
        acc = c_long()
        series = c_long()
        error = _dll.GetAcquisitionProgress(byref(acc), byref(series))
        if ERROR_CODE[error] == "DRV_SUCCESS":
            return acc.value
        else:
            return None

    def GetAcquiredDataNumpy(self, arr_ptr, array_dim):
        '''
        Returns the Acquired data

        Input:
            arr_ptr (ctypes.c_voird_p) : array pointer from numpy: data.ctypes.data_as(ctypes.c_void_p)
            array_dim (int) : total number of pixels

        Output:
            None
        '''
        error = _dll.GetAcquiredData(arr_ptr, array_dim)
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[error]

    def GetAcquiredData(self, imageArray):
        '''
        Returns the Acquired data

        Input:
            None

        Output:
            (array) : an array containing the acquired data
        '''
        if self._ReadMode == 0:
            dim = self._width
        elif self._ReadMode == 4:
            dim = self._width * self._height

        # ~ print "Dim is %s" % dim
        cimageArray = c_int * dim
        cimage = cimageArray()
        error = _dll.GetAcquiredData(pointer(cimage), dim)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

        for i in range(len(cimage)):
            imageArray.append(cimage[i])
        imageArray = imageArray[:]
        return imageArray

    def GetBitDepth(self):
        '''
        Returns the bit depth of the available channels

        Input:
            None

        Output:
            (int[]) : The bit depths
        '''
        bitDepth = c_int()
        self._bitDepths = []

        for i in range(self._noADChannels):
            _dll.GetBitDepth(i, byref(bitDepth))
            self._bitDepths.append(bitDepth.value)
        return self._bitDepths

    def GetPixelSize(self):
        '''
        Returns the camera pixel size
        Input:
            None

        Output:
            error (str)
            size (float, float) pixel size (width, height) in microns
            unsigned int WINAPI GetPixelSize(float* xSize, float* ySize)
        '''
        xsize = c_float()
        ysize = c_float()

        error = _dll.GetPixelSize(byref(xsize), byref(ysize))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return (ERROR_CODE[error], (xsize.value, ysize.value))

    def GetEMGainRange(self):
        '''
        Returns the number of completed accumulations

        Input:
            None

        Output:
            int) : The number of accumulations
        '''
        low = c_int()
        high = c_int()
        error = _dll.GetEMGainRange(byref(low), byref(high))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return low.value, high.value

    def GetNumberADChannels(self):
        '''
        Returns the number of AD channels

        Input:
            None

        Output:
            (int) : The number of AD channels
        '''
        noADChannels = c_int()
        error = _dll.GetNumberADChannels(byref(noADChannels))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return noADChannels.value

    def GetNumberPreAmpGains(self):
        '''
        Returns the number of Pre Amp Gains

        Input:
            None

        Output:
            (int) : The number of Pre Amp Gains
        '''
        noGains = c_int()
        error = _dll.GetNumberPreAmpGains(byref(noGains))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return noGains.value

    def GetSeriesProgress(self):
        '''
        Returns the number of completed kenetic scans

        Input:
            None

        Output:
            (int) : The number of completed kinetic scans
        '''
        acc = c_long()
        series = c_long()
        error = _dll.GetAcquisitionProgress(byref(acc), byref(series))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return series.value

    def GetStatus(self):
        '''
        Returns the status of the camera

        Input:
            None

        Output:
            (string) : DRV_IDLE
                       DRV_TEMPCYCLE
                       DRV_ACQUIRING
                       DRV_TIME_NOT_MET
                       DRV_KINETIC_TIME_NOT_MET
                       DRV_ERROR_ACK
                       DRV_ACQ_BUFFER
                       DRV_SPOOLERROR
        '''
        status = c_int()
        error = _dll.GetStatus(byref(status))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return ERROR_CODE[status.value]

    ###### Single Parameters Get/Set ######
    def GetEMCCDGain(self):
        '''
        Returns EMCCD Gain setting

        Input:
            None

        Output:
            (int) : EMCCD gain setting
        '''
        gain = c_int()
        error = _dll.GetEMCCDGain(byref(gain))
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return gain.value

    def SetEMCCDGain(self, gain):
        '''
        Set the EMCCD Gain setting

        Input:
            gain (int) : EMCCD setting

        Output:
            None
        '''
        error = _dll.SetEMCCDGain(gain)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def GetHSSpeed(self):
        '''
        Returns the available HS speeds of the selected channel

        Input:
            None

        Output:
            (float[]) : The speeds of the selected channel
        '''
        HSSpeed = c_float()
        HSSpeeds = []
        for i in range(self.GetNumberHSSpeeds()):
            error = _dll.GetHSSpeed(self._channel, self._outamp, i, byref(HSSpeed))
            if error != 20002:
                raise IOError(ERROR_CODE[error])
            HSSpeeds.append(HSSpeed.value)
        return HSSpeeds

    def SetHSSpeed(self, index):
        '''
        Set the HS speed to the mode corresponding to the index

        Input:
            index (int) : index corresponding to the Speed mode

        Output:
            None
        '''
        error = _dll.SetHSSpeed(index)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def GetVSSpeed(self):
        '''
        Returns the available VS speeds of the selected channel

        Input:
            None

        Output:
            (float[]) : The speeds of the selected channel
        '''
        VSSpeed = c_float()
        self._VSSpeeds = []

        for i in range(self.GetVSSpeed()):
            error = _dll.GetVSSpeed(i, byref(VSSpeed))
            if error != 20002:
                raise IOError(ERROR_CODE[error])
            self._VSSpeeds.append(VSSpeed.value)
        return self._VSSpeeds

    def SetVSSpeed(self, index):
        '''
        Set the VS speed to the mode corresponding to the index

        Input:
            index (int) : index corresponding to the Speed mode

        Output:
            None
        '''
        error = _dll.SetVSSpeed(index)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def GetPreAmpGain(self):
        '''
        Returns the available Pre Amp Gains

        Input:
            None

        Output:
            (float[]) : The pre amp gains
        '''
        gain = c_float()
        preAmpGain = []

        for i in range(self.GetNumberPreAmpGains()):
            error = _dll.GetPreAmpGain(i, byref(gain))
            if error != 20002:
                raise IOError(ERROR_CODE[error])
            preAmpGain.append(gain.value)
        return preAmpGain

    def SetPreAmpGain(self, index):
        '''
        Set the Pre Amp Gain to the mode corresponding to the index

        Input:
            index (int) : index corresponding to the Gain mode

        Output:
            None
        '''
        error = _dll.SetPreAmpGain(index)
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        self._preampgain = index

    ###### iDus interaction Functions ######
    def ShutDown(self):  # Careful with this one!!
        '''
        Shut down the Andor
        '''
        error = _dll.ShutDown()
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def AbortAcquisition(self):
        '''
        Abort the acquisition
        '''
        error = _dll.AbortAcquisition()
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def StartAcquisition(self):
        '''
        Start the acquisition
        '''
        error = _dll.StartAcquisition()
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetSingleImage(self):
        '''
        Shortcut to apply settings for a single scan full image
        '''
        self.SetReadMode(4)
        self.SetAcquisitionMode(1)
        print("Width: %d Height: %d" % (self._width, self._height))
        self.SetImage(1, 1, 1, self._width, 1, self._height)

    def SetSingleFVB(self):
        '''
        Shortcut to apply settings for a single scan FVB
        '''
        self.SetReadMode(0)
        self.SetAcquisitionMode(1)

    def GetAcquisitionTimings(self):
        '''
        Acquire all the relevant timings for acquisition,
        and store them in local memory
        
        returns:
            str  : error as a string
            dict : dict containing all real timings. Keys: 'exposure', 'accumulate', 'kinetic' 
        '''
        exposure = c_float()
        accumulate = c_float()
        kinetic = c_float()
        error = _dll.GetAcquisitionTimings(byref(exposure),
                                                byref(accumulate), byref(kinetic))
        self._exposure = exposure.value
        self._accumulate = accumulate.value
        self._kinetic = kinetic.value
        if error != 20002:
            raise IOError(ERROR_CODE[error])
        return (ERROR_CODE[error], dict(exposure=exposure.value, accumulate=accumulate.value, kinetic=kinetic.value))

    ###### Misc functions ######

    def SetShutter(self, typ, mode, closingtime=0, openingtime=10):
        '''
        Set the configuration for the shutter

        Input:
            typ         (int) : 0/1 Output TTL low/high signal to open shutter
            mode        (int) : 0/1/2 For Auto/Open/Close
            closingtime (int) : millisecs it takes to close
            openingtime (int) : millisecs it takes to open

        Output:
            None
        '''
        error = _dll.SetShutter(typ, mode, closingtime, openingtime)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetShutterEx(self, typ, mode, closingtime, openingtime, extmode):
        '''
        Set the configuration for the shutter in external mode

        Input:
            typ         (int) : 0/1 Output TTL low/high signal to open shutter
            mode        (int) : 0/1/2 For Auto/Open/Close
            closingtime (int) : millisecs it takes to close
            openingtime (int) : millisecs it takes to open
            extmode     (int) : 0/1/2 For Auto/Open/Close

        Output:
            None
        '''
        error = _dll.SetShutterEx(typ, mode, closingtime, openingtime,
                                       extmode)
        if error != 20002:
            raise IOError(ERROR_CODE[error])

    def SetSpool(self, active, method, path, framebuffersize):
        '''
        Set Spooling. Refer to manual for detailed description
        '''
        error = _dll.SetSpool(active, method, c_char_p(path), framebuffersize)
        if error != 20002:
            raise IOError(ERROR_CODE[error])


#####################################################

# List of error codes
ERROR_CODE = {
    20001: "DRV_ERROR_CODES",
    20002: "DRV_SUCCESS",
    20003: "DRV_VXNOTINSTALLED",
    20006: "DRV_ERROR_FILELOAD",
    20007: "DRV_ERROR_VXD_INIT",
    20010: "DRV_ERROR_PAGELOCK",
    20011: "DRV_ERROR_PAGE_UNLOCK",
    20013: "DRV_ERROR_ACK",
    20024: "DRV_NO_NEW_DATA",
    20026: "DRV_SPOOLERROR",
    20034: "DRV_TEMP_OFF",
    20035: "DRV_TEMP_NOT_STABILIZED",
    20036: "DRV_TEMP_STABILIZED",
    20037: "DRV_TEMP_NOT_REACHED",
    20038: "DRV_TEMP_OUT_RANGE",
    20039: "DRV_TEMP_NOT_SUPPORTED",
    20040: "DRV_TEMP_DRIFT",
    20050: "DRV_COF_NOTLOADED",
    20053: "DRV_FLEXERROR",
    20066: "DRV_P1INVALID",
    20067: "DRV_P2INVALID",
    20068: "DRV_P3INVALID",
    20069: "DRV_P4INVALID",
    20070: "DRV_INIERROR",
    20071: "DRV_COERROR",
    20072: "DRV_ACQUIRING",
    20073: "DRV_IDLE",
    20074: "DRV_TEMPCYCLE",
    20075: "DRV_NOT_INITIALIZED",
    20076: "DRV_P5INVALID",
    20077: "DRV_P6INVALID",
    20083: "P7_INVALID",
    20089: "DRV_USBERROR",
    20091: "DRV_NOT_SUPPORTED",
    20099: "DRV_BINNING_ERROR",
    20990: "DRV_NOCAMERA",
    20991: "DRV_NOT_SUPPORTED",
    20992: "DRV_NOT_AVAILABLE",
}

if __name__ == '__main__':
    A = AndorSDK()
    pass
